package com.flyfish.framework.r2dbc.metadata.impl;

import com.flyfish.framework.annotations.relation.Association;
import com.flyfish.framework.domain.base.Domain;
import com.flyfish.framework.r2dbc.metadata.R2dbcMetadataManager;
import com.flyfish.framework.r2dbc.metadata.R2dbcTableMetadata;
import com.flyfish.framework.r2dbc.metadata.reference.R2dbcAssociation;
import org.apache.commons.lang3.ClassUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.BeanUtils;
import org.springframework.core.annotation.AnnotatedElementUtils;
import org.springframework.data.mapping.model.Property;
import org.springframework.data.util.ClassTypeInformation;
import org.springframework.lang.NonNull;
import org.springframework.util.ReflectionUtils;

import java.beans.PropertyDescriptor;
import java.lang.reflect.Field;
import java.util.Arrays;
import java.util.Collection;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * 简单的元数据管理器
 *
 * @author wangyu
 * 基于线程安全的反射方法进行加载并持有
 */
public class SimpleR2dbcMetadataManager implements R2dbcMetadataManager {

    private final ConcurrentMap<Class<? extends Domain>, R2dbcTableMetadata> METADATA_STORE = new ConcurrentHashMap<>();

    /**
     * 获取元数据
     *
     * @param entityClass 实体类名
     * @return 结果
     */
    @Override
    public R2dbcTableMetadata getMetadata(Class<? extends Domain> entityClass) {
        return METADATA_STORE.computeIfAbsent(entityClass, this::computeMetadata);
    }

    /**
     * 计算元数据信息
     *
     * @param entityClass 实体类
     * @return 结果
     */
    private R2dbcTableMetadata computeMetadata(Class<? extends Domain> entityClass) {
        MetadataCreator creator = new MetadataCreator(entityClass);
        ReflectionUtils.doWithFields(entityClass, creator);
        return creator.metadata;
    }

    private static class MetadataCreator implements ReflectionUtils.FieldCallback {

        private final Map<String, PropertyDescriptor> descriptors;

        private final ClassTypeInformation<? extends Domain> typeInformation;

        private final R2dbcTableMetadata metadata;

        private MetadataCreator(Class<? extends Domain> entityClass) {
            this.descriptors = Arrays.stream(BeanUtils.getPropertyDescriptors(entityClass))
                    .collect(Collectors.toMap(PropertyDescriptor::getName, Function.identity()));
            this.typeInformation = ClassTypeInformation.from(entityClass);
            this.metadata = new R2dbcTableMetadata(entityClass);
        }

        /**
         * Perform an operation using the given field.
         *
         * @param field the field to operate on
         */
        @Override
        public void doWith(@NonNull Field field) throws IllegalArgumentException, IllegalAccessException {
            Property property = Property.of(typeInformation, field, descriptors.get(field.getName()));
            Class<?> fieldType = property.getType();

            ReflectionUtils.makeAccessible(field);

            Association association = AnnotatedElementUtils.findMergedAnnotation(field, Association.class);
            if (null != association) {
                if (ClassUtils.isAssignable(fieldType, Domain.class)) {
                    // 一对一关联
                    if (StringUtils.isNotBlank(association.field())) {
                        metadata.addAssociation(association.field(), true);
                    } else if (StringUtils.isNotBlank(association.foreignField())) {
                        metadata.addAssociation(association.foreignField(), false);
                    } else {
                       // 尚且不支持空白策略
                    }
                } else if (ClassUtils.isAssignable(fieldType, Collection.class)) {
                    // 一对多关联

                }
            }
        }
    }
}
